from hashlib import sha256

from aadhaar.exceptions import NumberOutOfRangeException


def generate_sha256_hexdigest(input_string: str, number_of_times: int) -> str:
    """
    Generate hex digest for n number of times if the number is between 2-9 else just once
    """
    if number_of_times not in range(0, 10):
        raise NumberOutOfRangeException('Number can be in range 0-9')
    digest_string = input_string
    if number_of_times == 0:
        number_of_times = 1
    for _ in range(number_of_times):
        digest_string = sha256(digest_string.encode('ISO-8859-1')).hexdigest()
    return digest_string



import os
import zipfile
from typing import Union
from operator import getitem
from functools import reduce
from collections import OrderedDict

import xmltodict

from aadhaar.utils import generate_sha256_hexdigest
from aadhaar.exceptions import EmptyArchiveException, NoXMLFileFound


class AadhaarXMLOffline:

    def __init__(self, path_to_aadhaar_zipfile: str, share_code: Union[str, int]) -> None:

        if not os.path.exists(path_to_aadhaar_zipfile):
            raise IOError("File doesn't exist")
        self.path_to_aadhaar_zipfile = path_to_aadhaar_zipfile
        self.share_code = str(share_code)

    def extract_data(self) -> OrderedDict:

        with zipfile.ZipFile(self.path_to_aadhaar_zipfile) as xml_zipfile:
            list_of_files = xml_zipfile.namelist()
            if not list_of_files:
                raise EmptyArchiveException('No files detected in the passed zip file')
            if not list_of_files[0].endswith('.xml'):
                raise NoXMLFileFound('No XML files found in the passed zip file')
            xml_file_name = list_of_files[0]
            with xml_zipfile.open(xml_file_name, pwd=str(self.share_code).encode()) as xml_file:
                xml_data = xml_file.read().decode()
        return xmltodict.parse(xml_data, attr_prefix='')

    def extract_reference_id(self) -> str:

        return self.extract_data()['OfflinePaperlessKyc']['referenceId']

    def validate_email(self, email: str) -> bool:

        traversal_list = ['OfflinePaperlessKyc', 'UidData', 'Poi', 'e']
        data = self.extract_data()
        email_hash = reduce(getitem, traversal_list, data)
        return email_hash == generate_sha256_hexdigest(str(email)+self.share_code, int(self.extract_reference_id()[3]))

    def validate_mobile_number(self, mobile_number: Union[str, int]) -> bool:

        traversal_list = ['OfflinePaperlessKyc', 'UidData', 'Poi', 'm']
        data = self.extract_data()
        email_hash = reduce(getitem, traversal_list, data)
        return email_hash == generate_sha256_hexdigest(
            str(mobile_number)+self.share_code,
            int(self.extract_reference_id()[3])
        )



import zlib
from io import BytesIO
from base64 import b64encode
from typing import Union, Dict

from PIL import Image

from aadhaar.utils import generate_sha256_hexdigest
from aadhaar.exceptions import MalformedIntegerReceived


class AadhaarSecureQR:
    """
    Decodes information from a Secure Aadhaar QR Data
    """

    BYTEORDER = 'big'
    ENCODING_USED = 'ISO-8859-1'
    _details = [
        "email_mobile_number_bit",
        "reference_id",
        "name",
        "dob",
        "gender",
        "care_of",
        "district",
        "landmark",
        "house",
        "location",
        "pincode",
        "post_office",
        "state",
        "street",
        "sub_district",
        "vtc",
    ]

    def __init__(self, integer_from_qr: Union[int, str]) -> None:

        if isinstance(integer_from_qr, str):
            try:
                integer_from_qr = int(integer_from_qr)
            except ValueError:
                raise TypeError('Passed string cannot be converted to an integer, please pass a valid integer')
        if not isinstance(integer_from_qr, int):
            raise TypeError('Expected integer value')

        self.integer_from_qr = integer_from_qr
        # Creating a byte sequence of 16 KB and removing null bytes(\x00) from the starting of the byte sequence

        self.byte_array = self.integer_from_qr.to_bytes(length=16*1024, byteorder=self.BYTEORDER).lstrip(b'\x00')

        # Keeping the window size to 16 + 15 -> 31 bits
        try:
            self.decompressed_byte_array = zlib.decompress(self.byte_array, wbits=16+zlib.MAX_WBITS)
        except zlib.error:
            raise MalformedIntegerReceived('Decompression failed, please send a valid integer received from QR code')

        self.decompressed_array_length = len(self.decompressed_byte_array)

        self.list_of_indexes_of_255_delimeter = [
            index
            for index in range(len(self.decompressed_byte_array))
            if self.decompressed_byte_array[index] == 255
        ]

        self._raw_extracted_data = {}

        previous = 0
        for detail, index_position in zip(self._details, self.list_of_indexes_of_255_delimeter):
            extracted_detail = self.decompressed_byte_array[previous:index_position].decode(self.ENCODING_USED)
            self._raw_extracted_data[detail] = extracted_detail
            previous = index_position + 1

    def is_mobile_present(self) -> bool:
        """
        from 0th index to the first 255 delimeter we get values from [0, 1, 2, 3] with following meaning
        0 -> email & mobile both are not available
        1 -> email is available but mobile isn't
        2 -> email isn't available but mobile is
        3 -> email and mobile both are available
        """

        return True if self._raw_extracted_data['email_mobile_number_bit'] in ('2', '3') else False

    def is_email_present(self) -> bool:
        """
        from 0th index to the first 255 delimeter we get values from [0, 1, 2, 3] with following meaning
        0 -> email & mobile both are not available
        1 -> email is available but mobile isn't
        2 -> email isn't available but mobile is
        3 -> email and mobile both are available
        """

        return True if self._raw_extracted_data['email_mobile_number_bit'] in ('1', '3') else False

    def get_signature(self) -> bytes:

        return self.decompressed_byte_array[self.decompressed_array_length-256:self.decompressed_array_length]

    def get_signed_data(self) -> bytes:

        return self.decompressed_byte_array[:self.decompressed_array_length-256]

    def get_mobile_sha256_hash(self) -> str:
        """returns sha256 hash value of the registered mobile number with aadhaar"""

        if self.is_mobile_present():
            return self.decompressed_byte_array[
                self.decompressed_array_length - 256 - 32: self.decompressed_array_length - 256
            ].hex()
        return ''

    def get_email_sha256_hash(self) -> str:
        """returns sha256 hash value of the email address registered with aadhaar"""

        if self.is_email_present():
            return self.decompressed_byte_array[
                self.decompressed_array_length - 256 - 32 - 32: self.decompressed_array_length - 256 - 32
            ].hex()
        return ''

    def get_image_data(self) -> str:
        """
        The compressed JPEG 2000 image is transformed into a base64 string for easy transportation (Rest APIs)
        """

        ending = self.decompressed_array_length - 256

        if self.is_mobile_present() and self.is_email_present():
            length_to_subtract = 32 * 2
        elif (self.is_mobile_present() and not self.is_email_present()) or (
            not self.is_mobile_present() and self.is_email_present()
        ):
            length_to_subtract = 32 * 1
        else:
            length_to_subtract = 32 * 0

        image_bytes = self.decompressed_byte_array[
            self.list_of_indexes_of_255_delimeter[15] + 1: ending - length_to_subtract
        ]

        img = Image.open(BytesIO(image_bytes))
        with BytesIO() as output:
            img.save(output, format='JPEG')
            image_data = output.getvalue()
        return b64encode(image_data).decode(self.ENCODING_USED)

    def validate_mobile_number(self, mobile_number: Union[str, int]) -> bool:
        """validates the input mobile number against the encoded mobile number embedded in the aadhaar qr code"""

        if not self.is_mobile_present():
            return False

        generated_hash = generate_sha256_hexdigest(
            str(mobile_number), int(self._raw_extracted_data['reference_id'][3])
        )
        return generated_hash == self.get_mobile_sha256_hash()

    def validate_email(self, email: str) -> bool:
        """validates the input email address against the encoded email address embedded in the aadhaar qr code"""

        if not self.is_email_present():
            return False

        generated_hash = generate_sha256_hexdigest(str(email), int(self._raw_extracted_data['reference_id'][3]))
        return generated_hash == self.get_email_sha256_hash()

    def extract_data(self) -> Dict[str, str]:
        """returns the extracted information from the qr code"""

        return {
            **self._raw_extracted_data,
            'photo': self.get_image_data(),
            'is_email_present': self.is_email_present(),
            'is_mobile_present': self.is_mobile_present(),
            'mobile_sha256_hash': self.get_mobile_sha256_hash(),
            'email_sha256_hash': self.get_email_sha256_hash(),
        }



import os
import zipfile
from unittest import TestCase
from tempfile import NamedTemporaryFile
from collections import OrderedDict

from aadhaar.qr import AadhaarSecureQR
from aadhaar.utils import generate_sha256_hexdigest
from aadhaar.exceptions import MalformedIntegerReceived, EmptyArchiveException, NoXMLFileFound
from aadhaar.offline_xml import AadhaarXMLOffline


class TestUtils(TestCase):

    def test_generate_sha256_hexdigest(self):

        test_hex_digest = 'd79ddb6ba12c0b0077571fadefe2cf3b875fc37a532ef6e084fcbef7e64e291a'
        self.assertEqual(generate_sha256_hexdigest('9999999999', 7), test_hex_digest)


class TestAadhaarSecureQR(TestCase):

    def setUp(self):

        self.sample_int = (
            "697941484820554848161929944287990190089397833259461440704476"
            "771748540728010407771465869816332540165921283092073423304757"
            "845470181056703201527022368291791582523470375471250488792130"
            "918178960780916888458384839645665300702247935633624019813036"
            "393088163236712473854151749949445813964737880868061416927322"
            "140474147659658395316924883137622439633516957706481298714057"
            "814488581947919017353764497023212514225396378497913801131879"
            "838544243609990162199828362481607008050483071259452576059693"
            "434157675562679159040363687813986166559938331942922836443418"
            "391319795873869700141049383928129869234282995156671253030975"
            "875936464970115363992197979842970756619926195003741817132928"
            "320737204801494866916066677619841404063338467710471769750752"
            "171758677670908420036495617886363610598886726092988757709295"
            "557040780378302139789734199991461679044102983722912974666922"
            "509563320109764432159350250340444071411051516703488912825896"
            "558343596503022584534856458205152134880074257444287708777419"
            "466898351662963107334120270545338278061377542733694928338808"
            "489165448422544694094166094244063778474429325991647984140708"
            "818946296448967023186648190423733849487281309889087584564003"
            "437037038710879895018022086543601275248721667704181731293011"
            "974760101780757756541397754569337548013132424069609987947943"
            "672257656644793959319559068459126180903802312217817200615049"
            "956918521883874933723828159703728892446400999753093833679817"
            "602359729232832096508699018453142618886296540831330897349592"
            "496514411339659382909064526665331377458203613898201336856147"
            "471915444713489446661156058975825182906322637030028217582347"
            "956984726143934840455825140227373086505348221458918002830204"
            "382143835758330281837414397399700274504752640575576040704500"
            "669442350133708178029981508032484033782881264430004190035681"
            "642911426109823019897675202600207987688279659723561501559448"
            "618205778147615291817074640315700521689623942852170603346606"
            "158760806503613315307443219595213136856423416800544777019034"
            "577702491762987963917116171992985207826530916075926098959061"
            "815888989183529473561436667450396158444549768573631262824848"
            "355198652986742301625547655369192205424168623096897522951170"
            "092817128154990268236530233367741295178883980686979604051223"
            "589931173433785868453115672141628011447336882646309848525239"
            "426007579038641587529092257056868643958603626246541400233411"
            "787008892280166052941475978431879984380613009699819088124040"
            "413886929330978233530529672066622024330417508635827821135578"
            "995799801480120933229345894046385910659198643452043381058356"
            "930922492926422826384147737894932931244395821593929443266946"
            "426021653407456088272300683845979281234025307833029113552695"
            "267520379083343023785283174060143319836424336356973020535107"
            "739344169114124005590081909122993160514686552018300181023970"
            "846432258838995603629176017555884381910541823458023961017432"
            "363660609526272294014370606369884649967328537762118057053778"
            "816030493680991523788948934238789105701278372669492018457320"
            "278967296392238002827112444802426564439668634150844783035138"
            "024212754239384941028383040959498850324679954444468760695488"
            "151059751568641099382890758897969914118016089306260333810485"
            "790323984585678313027593541356927543990878998331166321193744"
            "9259444259898972766208"
        )
        self.secure_qr = AadhaarSecureQR(self.sample_int)

    def test_b64encoded_image(self):

        image_b64_data = (
            '/9j/4AAQSkZJRgABAQAAAQABAAD/2wBDAAgGBgcGBQgHBwcJCQgKDBQNDAsLDBkSEw8UHRofHh0aHBwgJC4nICI'
            'sIxwcKDcpLDAxNDQ0Hyc5PTgyPC4zNDL/2wBDAQkJCQwLDBgNDRgyIRwhMjIyMjIyMjIyMjIyMjIyMjIyMjIyMj'
            'IyMjIyMjIyMjIyMjIyMjIyMjIyMjIyMjIyMjL/wAARCAA8ADwDASIAAhEBAxEB/8QAHwAAAQUBAQEBAQEAAAAAA'
            'AAAAAECAwQFBgcICQoL/8QAtRAAAgEDAwIEAwUFBAQAAAF9AQIDAAQRBRIhMUEGE1FhByJxFDKBkaEII0KxwRVS'
            '0fAkM2JyggkKFhcYGRolJicoKSo0NTY3ODk6Q0RFRkdISUpTVFVWV1hZWmNkZWZnaGlqc3R1dnd4eXqDhIWGh4i'
            'JipKTlJWWl5iZmqKjpKWmp6ipqrKztLW2t7i5usLDxMXGx8jJytLT1NXW19jZ2uHi4+Tl5ufo6erx8vP09fb3+P'
            'n6/8QAHwEAAwEBAQEBAQEBAQAAAAAAAAECAwQFBgcICQoL/8QAtREAAgECBAQDBAcFBAQAAQJ3AAECAxEEBSExB'
            'hJBUQdhcRMiMoEIFEKRobHBCSMzUvAVYnLRChYkNOEl8RcYGRomJygpKjU2Nzg5OkNERUZHSElKU1RVVldYWVpj'
            'ZGVmZ2hpanN0dXZ3eHl6goOEhYaHiImKkpOUlZaXmJmaoqOkpaanqKmqsrO0tba3uLm6wsPExcbHyMnK0tPU1db'
            'X2Nna4uPk5ebn6Onq8vP09fb3+Pn6/9oADAMBAAIRAxEAPwD1S/8AGdjb+HJdSjwZl+RYSeRIRwPp3z6A9+K8tn'
            '1G41q5e7uZW+Y5A9K53xLqkv2myhDMkbAkr75xWlBOFjRSygY7mpc7RLhC7uy8sUX3tgJ96iuAGH3Rj0phvIY+r'
            'p+BqJ76BhkNXM+Z6nZFLoU54VwcRgCmWX2V5GtruMPDKNpz2FOnvIyCAy49zVMSRysAkiBu3zVKTvc0k1azKUYn'
            '8F+J02qJrVyXj3k7ZIzzg47j+lem29888CTKqYcA5Hy598CuB8Qyef4ejjmT97bOGRu+OhFWtCvpp9Ih3yOfLGx'
            'ceg6V0x1RxSSuZ/iiCV5rGZxkCQRtjtk8Vl3s8jTMHlZQOFC1sXM8lxcRW8gIBw5PoQeKdcaXHNkKoDetQ5pM2h'
            'TfKckXm8whZpMr1BrbsIJry0crK64H3h1qc6G5d5XfryxHetfTLX7PZSbVABFRKqnsa06bT1OG1GSZpTGsjBvbv'
            'VO3A84oZnDjgmulu7CN5ww4eootIddwBBDnJzRGStqTODbFimnk0S7inkL+Ug2N7Z6V6T4N8Nvb+F7Q3SASygyE'
            'egJ4H5YriI7JbfTp0dcgr2+or2bwvfR6/oEF7HAsfWNowQdhU4x/I/QirTMZU3c8mvh5ZEi/eIH5c1FFenfk8+1'
            'XtTtnFsJeNo4rnCzeZhSR71lKOp005+6bU90fJYqg4GSKxk8RSlXQK0aduKZLqccAMZck96ovewN/yxmb6/8A6q'
            'IxKc30LttqKT5zH+8U4DY61feddue9c+t2kZYx5HPIIxj6VcgfzRuHNOUAU1bU2vOZtOlIGWA6fiK9N+Gen3Z8K'
            'NILh40kupCijpgYU/qDXnOkaVqGqxPFp1q1xICCwDqu3t1YgdeK900GwGm6FZWRVY2hhVGCdC2OT+eaaRlKep5Z'
            'PbLcQNC/3TXEX0MlnMYpAVkUHHoRXfD74rmvHahILN1GGywz/wB81c1qYU5PY5OJTuLrGCx7mkmW+PKrhRT7aRx'
            '0arTzyeWTuNRex2RWhkvGXH7xBuHer9kPKh5/DNQMS5yxyat6cxOp2kZPyGVcj8ad7mUl7x7T4A0SXSNH+0XSEX'
            'F185jdcFF7Ag9D1J+vtXXFzn09qr25LIrHqRk1NVIwm9T/2Q=='
        )
        self.assertEqual(self.secure_qr.get_image_data(), image_b64_data)

    def test_mobile_presence(self):

        original_bit = self.secure_qr._raw_extracted_data['email_mobile_number_bit']
        self.secure_qr._raw_extracted_data['email_mobile_number_bit'] = '0'
        self.assertFalse(self.secure_qr.is_mobile_present())
        self.secure_qr._raw_extracted_data['email_mobile_number_bit'] = '1'
        self.assertFalse(self.secure_qr.is_mobile_present())
        self.secure_qr._raw_extracted_data['email_mobile_number_bit'] = '2'
        self.assertTrue(self.secure_qr.is_mobile_present())
        self.secure_qr._raw_extracted_data['email_mobile_number_bit'] = '3'
        self.assertTrue(self.secure_qr.is_mobile_present())
        self.secure_qr._raw_extracted_data['email_mobile_number_bit'] = original_bit

    def test_email_presence(self):

        original_bit = self.secure_qr._raw_extracted_data['email_mobile_number_bit']
        self.secure_qr._raw_extracted_data['email_mobile_number_bit'] = '0'
        self.assertFalse(self.secure_qr.is_email_present())
        self.secure_qr._raw_extracted_data['email_mobile_number_bit'] = '1'
        self.assertTrue(self.secure_qr.is_email_present())
        self.secure_qr._raw_extracted_data['email_mobile_number_bit'] = '2'
        self.assertFalse(self.secure_qr.is_email_present())
        self.secure_qr._raw_extracted_data['email_mobile_number_bit'] = '3'
        self.assertTrue(self.secure_qr.is_email_present())
        self.secure_qr._raw_extracted_data['email_mobile_number_bit'] = original_bit


class TestExceptionsOnAadhaarSecureQR(TestCase):

    def test_raise_error_on_unaccepted_type(self):

        for type_ in [{'k': 'v'}, [123], 123.4123, 'abc']:
            with self.assertRaises(TypeError):
                AadhaarSecureQR(type_)

    def test_raise_error_on_integer_that_cant_be_decompressed(self):

        with self.assertRaises(MalformedIntegerReceived):
            AadhaarSecureQR(23123891283801283010283091820930910923809182038019280380128309)


class TestAadhaarOfflineXML(TestCase):

    def setUp(self):

        self.BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        self.RESOURCES_DIR = os.path.join(self.BASE_DIR, 'resources')
        self.zip_file_path = os.path.join(self.RESOURCES_DIR, 'test.zip')
        self.SHARE_CODE = 1234

    def test_working_aadhaar_offline_xml(self):

        xml_offline = AadhaarXMLOffline(self.zip_file_path, self.SHARE_CODE)
        decoded_aadhaar_data = xml_offline.extract_data()
        self.assertIn('OfflinePaperlessKyc', decoded_aadhaar_data)
        self.assertIn('referenceId', decoded_aadhaar_data['OfflinePaperlessKyc'])
        self.assertIn('UidData', decoded_aadhaar_data['OfflinePaperlessKyc'])
        self.assertIn('Poi', decoded_aadhaar_data['OfflinePaperlessKyc']['UidData'])
        self.assertIn('Poa', decoded_aadhaar_data['OfflinePaperlessKyc']['UidData'])
        self.assertIn('Pht', decoded_aadhaar_data['OfflinePaperlessKyc']['UidData'])
        self.assertIsInstance(decoded_aadhaar_data, OrderedDict)

    def test_not_zipfile(self):

        with NamedTemporaryFile(mode='w') as ntf:
            ntf.write('Sometext')
            with self.assertRaises(zipfile.BadZipfile):
                xml_offline = AadhaarXMLOffline(ntf.name, self.SHARE_CODE)
                xml_offline.extract_data()

    def test_empty_zipfile(self):

        file_name = 'empty.zip'
        with open(file_name, 'wb') as empty_zipfile:
            empty_zipfile.write(b'PK\x05\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
        with self.assertRaises(EmptyArchiveException):
            xml_offline = AadhaarXMLOffline(file_name, self.SHARE_CODE)
            xml_offline.extract_data()
        os.remove(file_name)

    def test_no_xml_zipfile(self):

        zipfile_name = 'no_xml.zip'
        with zipfile.ZipFile(zipfile_name, 'w') as zf:
            zf.writestr('test.txt', 'Sample Text Data')
        with self.assertRaises(NoXMLFileFound):
            xml_offline = AadhaarXMLOffline(zipfile_name, self.SHARE_CODE)
            xml_offline.extract_data()
        os.remove(zipfile_name)

    def test_invalid_sharecode(self):

        with self.assertRaises(RuntimeError):
            xml_offline = AadhaarXMLOffline(self.zip_file_path, 3456)
            xml_offline.extract_data()



class NumberOutOfRangeException(Exception):

    pass


class MalformedIntegerReceived(Exception):

    pass


class EmptyArchiveException(Exception):

    pass


class NoXMLFileFound(Exception):

    pass
